unit gfgltree;

interface

uses classes, comctrls, controls, messages{, DbugIntf};

{
TgfglTreeView : recursing stateimages treeview (for checkboxes)
FREE delphi component D2/3/4/5
copyright Greg Lorriman 1999 greg@lorriman.demon.co.uk
http://www.lorriman.demon.co.uk

Bug reports/fixes welcomed
So is love and attention. If you like this component then tell me - I won't be
angry :)  greg@lorriman.demon.co.uk

Description :
   Recurses (propogates) the "state" of a node (ie no-tick,tick,greytick).
   This is useful for checkbox treeviews. Stateimages are user defined via
   an imageList (handy images are supplied in the archive).

Features :
      Design time functionality (important : read instructions before trying)
      State Persistence (statesSaveToFile etc)
      Recursive behaviour supported via normal stateindex property of nodes.
      Branch can be excluded from recursion by setting stateindex of parent to
      value of StateIgnore property (default 0).
      This help is available via 'Help' property.

Known Limitations :

   No facility for "conditional exclusion".
   No speed or persistence size optimisation

Basic Usage :

   a)Set the stateimages property to an imagelist with your chosen state images.
   b)The items in the imagelist should be ordered as follows :
      0=anything (use 'empty' for convenience)
      1=empty
      2=ticked
      3=grey tick
   When adding icons to imagelist in D2 use 'stretch' option
   c)Set StateRecursion:=true;
   d)add some nodes (leave their stateindex at -1)
   d)run and click on node state images to test

Programmatic usage :

   You can set the state of a node (and get the same affect as clicking) via
   the property "stateIndex" of each node.
   You can "re-sync" the states by toggling the StateRecursion property.

Design time usage :

   The standard node editor does some strange things so design time usage
   works as follows :

   Valid states are 2 (ticked) and StateIgnore (ignore branch, default=0).
   Do not use other states as they will be ignored.
   The states are recursed at run-time (if StateRecusion:=true).
   It'll look a bit messy at design time. Empty and grey ticks will appear in
   the appropriate places at run time.

Persistence :

   Streaming can be 'chained' with streaming of other info.

   Saved states are compatible only with tree exactly as when saved. You'll
   only get an exception when loading if there are more nodes than original and
   stream falls off end.

   When loading it is necessary to set StateRecursion to false and then back to
   true after load. This allows these methods to be usable even if you are
   not using state recursion.

   You can probabaly get a speed up by invoking beginupdate/endupdate.

   Persistence is StateIgnore aware. Temporarily Change StateIgnore to an
   unused value if you wish to load/save ignored branches. Loading does not
   write to the stateindex of the root of an ignored branch.

   Persistence does not save the StateIgnore Value to the stream. If you need
   to do this then use the stream methods (rather than file) and write/read
   the value before calling the state-streaming method.

Notes :

Use StateIgnore to define the state that blocks the algorithm (ignores node
and it's children). Default=0. Changing StateIgnore does not cause an update
of the tree. Do this by toggling the StateRecursion property.

Clicking a stateimage can cause collapse and expansion. Optionally use this
code to prevent (in the appropriate events) :

   allowcollapse:=not (TgfglTreeView(sender).MouseOnStateImage);

and same again for the expand event. Why isn't this incorporated? Because it
would impose a change to the default behaviour of the treeview component.

SCROLL TO TOP FOR START OF HELP
//ignore message above (for property help generator)
}

type
   TgfglTreeView=class(TTreeView)

   private

      FStateRecursion : boolean;
      FHelp: TStrings;
      FStateIgnore: integer;

      procedure propogateStateDown(node : TTreeNode);//usually child
      procedure propogateStateUp(node : TTreeNode);//usually parent
      procedure stateSaveToStream(node: TTreeNode; stream: Tstream);
      procedure stateLoadFromStream(node: TTreeNode; stream: TStream);

      procedure SetVersion(const Value: string);
      function GetVersion: string;
      function getOtherstuff: string;
      procedure setOtherStuff(const Value: string);
      procedure setHelp(const Value: TStrings);
      procedure SetStateIgnore(const Value: integer);

   protected

    //for recursing the state change of a node
      procedure propogateState(node : TTreeNode);virtual;

    //for component editor support - normally items.getfirstnode
      procedure syncState(node : TTreeNode);virtual;

    //general
      procedure setStateRecursion(const Value : boolean); virtual;
      procedure wndproc(var Message : TMessage); override;
      procedure loaded; override;
      procedure nodeStateChangeNotify(node : TTreeNode);virtual;

    //user interaction support
      procedure MouseDown(Button : TMouseButton; Shift : TShiftState;
         X, Y : Integer); override;

   public

      constructor create(AOwner: Tcomponent);override;
      destructor destroy;override;

   //persistence
      procedure statesSaveToStream(stream : Tstream);virtual;
      procedure statesLoadFromStream(stream : TStream);virtual;
      procedure StatesSaveToFile(filename : string);
      procedure StatesLoadFromFile(filename : string);

   //for user to stop expand/collapse if required
      function mouseOnStateImage : boolean;

   published

      property StateRecursion : boolean read FStateRecursion write SetStateRecursion;
      property StateIgnore : integer read FStateIgnore write SetStateIgnore;

      property Version : string read GetVersion write SetVersion stored false;
      property OtherStuff : string read getOtherstuff write setOtherStuff stored false;
      property Help : TStrings  read FHelp write setHelp stored False;

   end;

procedure Register;

implementation

uses windows, commctrl, sysutils;


constructor TgfglTreeView.create(AOwner: Tcomponent);
const helptext : string=   'TgfglTreeView : recursing stateimages treeview (for checkboxes)'#13#10+
                           'FREE delphi component D2/3/4/5'#13#10+
                           'copyright Greg Lorriman 1999 greg@lorriman.demon.co.uk'#13#10+
                           'http://www.lorriman.demon.co.uk'#13#10+
                           ''#13#10+
                           'Bug reports/fixes welcomed'#13#10+
                           'So is love and attention. If you like this component then tell me - I won'#39't be'#13#10+
                           'angry :)  greg@lorriman.demon.co.uk'#13#10+
                           ''#13#10+
                           'Description :'#13#10+
                           '   Recurses (propogates) the "state" of a node (ie no-tick,tick,greytick).'#13#10+
                           '   This is useful for checkbox treeviews. Stateimages are user defined via'#13#10+
                           '   an imageList (handy images are supplied in the archive).'#13#10+
                           ''#13#10+
                           'Features :'#13#10+
                           '      Design time functionality (important : read instructions before trying)'#13#10+
                           '      State Persistence (statesSaveToFile etc)'#13#10+
                           '      Recursive behaviour supported via normal stateindex property of nodes.'#13#10+
                           '      Branch can be excluded from recursion by setting stateindex of parent to'#13#10+
                           '      value of StateIgnore property (default 0).'#13#10+
                           '      This help is available via '#39'Help'#39' property.'#13#10+
                           ''#13#10+
                           'Known Limitations :'#13#10+
                           ''#13#10+
                           '   No facility for "conditional exclusion".'#13#10+
                           '   No speed or persistence size optimisation'#13#10+
                           ''#13#10+
                           'Basic Usage :'#13#10+
                           ''#13#10+
                           '   a)Set the stateimages property to an imagelist with your chosen state images.'#13#10+
                           '   b)The items in the imagelist should be ordered as follows :'#13#10+
                           '      0=anything (use '#39'empty'#39' for convenience)'#13#10+
                           '      1=empty'#13#10+
                           '      2=ticked'#13#10+
                           '      3=grey tick'#13#10+
                           '   When adding icons to imagelist in D2 use '#39'stretch'#39' option'#13#10+
                           '   c)Set StateRecursion:=true;'#13#10+
                           '   d)add some nodes (leave their stateindex at -1)'#13#10+
                           '   d)run and click on node state images to test'#13#10+
                           ''#13#10+
                           'Programmatic usage :'#13#10+
                           ''#13#10+
                           '   You can set the state of a node (and get the same affect as clicking) via'#13#10+
                           '   the property "stateIndex" of each node.'#13#10+
                           '   You can "re-sync" the states by toggling the StateRecursion property.'#13#10+
                           ''#13#10+
                           'Design time usage :'#13#10+
                           ''#13#10+
                           '   The standard node editor does some strange things so design time usage'#13#10+
                           '   works as follows :'#13#10+
                           ''#13#10+
                           '   Valid states are 2 (ticked) and StateIgnore (ignore branch, default=0).'#13#10+
                           '   Do not use other states as they will be ignored.'#13#10+
                           '   The states are recursed at run-time (if StateRecusion:=true).'#13#10+
                           '   It'#39'll look a bit messy at design time. Empty and grey ticks will appear in'#13#10+
                           '   the appropriate places at run time.'#13#10+
                           ''#13#10+
                           'Persistence :'#13#10+
                           ''#13#10+
                           '   Streaming can be '#39'chained'#39' with streaming of other info.'#13#10+
                           ''#13#10+
                           '   Saved states are compatible only with tree exactly as when saved. You'#39'll'#13#10+
                           '   only get an exception when loading if there are more nodes than original and'#13#10+
                           '   stream falls off end.'#13#10+
                           ''#13#10+
                           '   When loading it is necessary to set StateRecursion to false and then back to'#13#10+
                           '   true after load. This allows these methods to be usable even if you are'#13#10+
                           '   not using state recursion.'#13#10+
                           ''#13#10+
                           '   You can probabaly get a speed up by invoking beginupdate/endupdate.'#13#10+
                           ''#13#10+
                           '   Persistence is StateIgnore aware. Temporarily Change StateIgnore to an'#13#10+
                           '   unused value if you wish to load/save ignored branches. Loading does not'#13#10+
                           '   write to the stateindex of the root of an ignored branch.'#13#10+
                           ''#13#10+
                           '   Persistence does not save the StateIgnore Value to the stream. If you need'#13#10+
                           '   to do this then use the stream methods (rather than file) and write/read'#13#10+
                           '   the value before calling the state-streaming method.'#13#10+
                           ''#13#10+
                           'Notes :'#13#10+
                           ''#13#10+
                           'Use StateIgnore to define the state that blocks the algorithm (ignores node'#13#10+
                           'and it'#39's children). Default=0. Changing StateIgnore does not cause an update'#13#10+
                           'of the tree. Do this by toggling the StateRecursion property.'#13#10+
                           ''#13#10+
                           'Clicking a stateimage can cause collapse and expansion. Optionally use this'#13#10+
                           'code to prevent (in the appropriate events) :'#13#10+
                           ''#13#10+
                           '   allowcollapse:=not (TgfglTreeView(sender).MouseOnStateImage);'#13#10+
                           ''#13#10+
                           'and same again for the expand event. Why isn'#39't this incorporated? Because it'#13#10+
                           'would impose a change to the default behaviour of the treeview component.'#13#10+
                           ''#13#10+
                           'SCROLL TO TOP FOR START OF HELP';



begin
   inherited create(AOwner);
   FHelp:=TStringlist.create;
   FHelp.text:=helptext;
end;

destructor TgfglTreeView.destroy;
begin
   fhelp.free;
   inherited;
end;


procedure TgfglTreeView.SetStateRecursion(const Value : boolean);
begin
   if Value<>FStateRecursion then begin
      FStateRecursion:=Value;
      if value then
         syncState(items.getfirstnode);
   end;
end;

procedure TgfglTreeView.nodeStateChangeNotify(node: TTreeNode);
begin
   propogateState(node);
end;

procedure TgfglTreeView.MouseDown(Button : TMouseButton; Shift : TShiftState;
   X, Y : Integer);
var
   index : integer;
begin
   inherited;

   if not FStateRecursion then exit;

   if assigned(selected)then begin
      if(htonstateicon in gethittestinfoat(x, y))then begin
         index:=selected.stateindex;
         inc(index);
         if index>2 then
            index:=1;
         selected.stateindex:=index;
      end;
   end;
end;

function TgfglTreeView.MouseOnStateImage : boolean;
var
   mousepos : Tpoint;
begin
   getcursorpos(mousepos);
   mousepos:=screentoclient(mousepos);
   result:=htonstateicon in gethittestinfoat(mousepos.x, mousepos.y);
end;

procedure TgfglTreeView.wndproc(var Message : TMessage);
var
   node : TTreeNode;
begin
   inherited;
   if([csDesigning, csLoading, csReading]*componentstate)<>[] then exit;
   if not FStateRecursion then exit;

   if message.msg=TVM_SETITEM then begin
      if(PTVITEM(message.lparam).mask and(TVIF_STATE or TVIF_HANDLE))>0 then begin
         node:=items.getnode(PTVITEM(message.lparam).hitem);
         if node.stateindex in [1, 2] then
            nodeStateChangeNotify(node);
      end;
   end;
end;


procedure TgfglTreeView.propogateState(node : TTreeNode);
begin
   FStateRecursion:=False;
   try
      propogateStateDown(node);
      if assigned(node.parent)then
         propogateStateUp(node.parent);
   finally
      FStateRecursion:=true;
   end;
end;

procedure TgfglTreeView.propogateStateDown(node : TTreeNode);
var
   child : TTreeNode;
begin
   with node do begin
      if stateindex=FStateIgnore then exit;

      child:=getfirstchild;
      while assigned(child) do begin
         if child.stateindex<>stateindex then begin
            if child.stateindex<>FStateIgnore then
               child.stateindex:=stateindex;
            propogateStateDown(child);
         end;
         child:=getnextchild(child);
      end;
   end;
end;

procedure TgfglTreeView.propogateStateUp(node : TTreeNode);
var
   flag : integer;
   child : TTreeNode;
begin
   with node do begin

      if stateindex=FStateIgnore then exit;

      flag:=0;

      child:=getfirstchild;
      while assigned(child) do begin
         case child.stateindex of
            -1 : flag:=flag or 1;
            1  : flag:=flag or 1;
            2  : flag:=flag or 2;
            3  : flag:=flag or 4;
         end;
         child:=getnextchild(child);
      end;

      case flag of
         1 : stateindex:=1;
         2 : stateindex:=2;
      else stateindex:=3;
      end;

      if assigned(parent)then
         propogateStateUp(parent);
   end;
end;


procedure TgfglTreeView.loaded;
begin
   inherited;
   if csDesigning in componentstate then exit;

   if FStateRecursion then
      syncState(items.getfirstnode);
end;



procedure TgfglTreeView.syncState(node : TTreeNode);
var
   child : TTreeNode;
begin
   while assigned(node)do begin
      if node.stateindex<>FStateIgnore then
      with node do begin
         if stateindex=2 then
            propogateState(node)
         else begin
            if stateindex=-1 then begin
               FStateRecursion:=false;
               stateindex:=1;
               FStateRecursion:=true;
            end;
            child:=getfirstchild;
            while assigned(child) do begin
               if child.stateindex<>FStateIgnore then
                  syncState(child);
               child:=getnextchild(child);
            end;
         end;
      end;
      node:=node.getNextSibling;
   end;
end;

procedure TgfglTreeView.stateLoadFromStream(node : TTreeNode;stream: TStream);
var
   state : integer;
   child : TTreeNode;
begin
   while assigned(node) do begin
      with node do begin
         stream.read(state,sizeof(state));
         if state=StateIgnore then
            exit;
         stateIndex:=state;
         child:=getfirstchild;
         while assigned(child) do begin
            stateLoadFromStream(child,stream);
            child:=getnextchild(child);
         end;
      end;
      node:=node.getNextSibling;
   end;
end;

procedure TgfglTreeView.stateSaveToStream(node : TTreeNode;stream: Tstream);
var
   state : integer;
   child : TTreeNode;
begin
   while assigned(node) do begin
      with node do begin
         state:=stateIndex;
         stream.write(state,sizeof(state));
         if state=StateIgnore then
            exit;
         child:=getfirstchild;
         while assigned(child) do begin
            stateSaveToStream(child,stream);
            child:=getnextchild(child);
         end;
      end;
      node:=node.getNextSibling;
   end;
end;


procedure TgfglTreeView.statesSaveToStream(stream: Tstream);
begin
   stateSaveToStream(items.getfirstnode,stream);
end;

procedure TgfglTreeView.statesLoadFromStream(stream: TStream);
begin
   StateLoadFromStream(items.getfirstnode,stream);
end;

procedure TgfglTreeView.StatesLoadFromFile(filename: string);
var
   stream : TFileStream;
begin
   stream:=TFileStream.create(filename,fmShareDenyNone + fmOpenRead);
   try
      statesLoadFromStream(stream);
   finally
      stream.free;
   end;
end;

procedure TgfglTreeView.StatesSaveToFile(filename: string);
var
   stream : TFileStream;
begin
   stream:=TFileStream.create(filename,fmCreate);
   try
      statesSaveToStream(stream);
   finally
      stream.free;
   end;
end;

function TgfglTreeView.getOtherstuff: string;
begin
   result:='http://www.lorriman.demon.co.uk';
end;

function TgfglTreeView.GetVersion: string;
begin
   result:='.1 beta';
end;

procedure TgfglTreeView.setHelp(const Value: TStrings);
begin

end;

procedure TgfglTreeView.setOtherStuff(const Value: string);
begin

end;

procedure TgfglTreeView.SetVersion(const Value: string);
begin

end;

procedure TgfglTreeView.SetStateIgnore(const Value: integer);
begin
  FStateIgnore := Value;
end;


procedure Register;
begin
   RegisterComponents({$IFDEF VER90}'Win95'{$ELSE}'Win32'{$ENDIF}, [TgfglTreeView]);
end;


end.

